/***************************************************************
 * Name:      WDGComm.cpp
 * Purpose:   Watchdog communication class
 * Author:    Alexey Govorukhin (MasterAlexei@googlemail.com)
 * Created:   2010
 * Copyright: Alexey Govorukhin (http://www.fun-electronic.net/)
 * License:
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the author nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 **************************************************************************/
#if !defined(_WDGCOMM_C)
#define _WDGCOMM_C

#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>

#include <stdio.h>
#include <string.h>


#include "types.h"
#include "Config.h"

#include "WDGProtocol.h"
#include "SoketComm.h"
#include "Logging.h"

#include "Watchdog.h"

#include "WdgCommProtocol.h"
#include "WDGComm.h"

#include "GlobalData.h"
#include "fcs.h"
#include "TimerDefs.h"
#include "TimerClass.h"

#include "TaskDefs.h"
#include "tasks.h"

#include "Externals.h"


/**************************************************************************
* definitions
**************************************************************************/


/**************************************************************************
* declarations
**************************************************************************/

TWDConnect * WDConnect = NULL;

/**************************************************************************
* function prototypes
**************************************************************************/

static int gsPortDsc = -1;

/**************************************************************************
* implementation
**************************************************************************/
void signal_handler_IO ( int status )
{
  u08 bByte;
  int res = 0;

  while ( (res = read( gsPortDsc, &bByte, 1 )) > 0 )
  {
    if( WDConnect != NULL )
    {
      WDConnect->ReceiveByte( bByte );
    }
  }
}

void TWDConnect::ReceiveByte( u08 bByte )
{
  u16convert wCRC;

  bRxBuff[ wRxIndex ] = bByte;

  if( bRxBuff[ 0 ] == UART_SYNC_BYTE )
  {
    wRxIndex++;
    if( wRxIndex >= sizeof(TWD_CommData) + 2 )
    {
      LOG( LOG_DEBUG, "%u bytes received from WDG", wRxIndex );

      wCRC.value = CalculateCRC( bRxBuff, sizeof( TWD_CommData ) );
      if (
            ( wCRC.bytes.low  == bRxBuff[ sizeof( TWD_CommData )    ] ) &&
            ( wCRC.bytes.high == bRxBuff[ sizeof( TWD_CommData ) + 1] )
         )
      {
        if( Timer != NULL )
        {
          Timer->Clear(  SWT_ANSWER_DELAY_TIMER );
        }
        nPortNum = 0;

        ProcessData( (TWD_CommData *)bRxBuff );
      }
      else
      {
        LOG( LOG_DEBUG, "BAD CRC" );
      }

      wRxIndex = 0;
      bRxBuff[0] = 0;
    }
  }
  else
  {
    wRxIndex = 0;
  }
}


TWDConnect::TWDConnect( char * szPrtNm )
{
  fOpened = false;
  wRxIndex = 0;

  nPNLen = strlen( szPrtNm );

  pszPortName = new char[ nPNLen + 3 ];
  if( pszPortName != NULL )
  {
    memset( pszPortName, 0, nPNLen + 3 );
    strcpy( pszPortName, szPrtNm );

  }
  nPortNum = 0;
}

TWDConnect::~TWDConnect()
{
  Close();

  if( pszPortName != NULL )
  {
    delete [] pszPortName;
  }
}

void TWDConnect::Close()
{
  if( fOpened )
  {
    // restore old port settings
    tcsetattr( PortDsc, TCSANOW, &oldtio );
    close( PortDsc );
  }
  fOpened = false;
}

int TWDConnect::Open()
{
  int nRes = -1;
  struct sigaction saio;

  Close();

  pszPortName[ nPNLen ] = nPortNum + '0';

  PortDsc = open( pszPortName, (O_RDWR | O_NOCTTY | O_NONBLOCK) );   // open the device to be non-blocking (read will return immediatly)

  if ( PortDsc >= 0 )
  {
    gsPortDsc = PortDsc;
                                                                    // install the signal handler before making the device asynchronous
    sigemptyset(&saio.sa_mask);
    saio.sa_handler   = signal_handler_IO;
    saio.sa_flags     = 0;
    saio.sa_restorer  = NULL;
    if( 0 == sigaction( SIGIO, &saio, NULL) )
    {

      // allow the process to receive SIGIO
      fcntl( PortDsc, F_SETOWN, getpid() );

      // Make the file descriptor asynchronous
      fcntl( PortDsc, F_SETFL, FASYNC );

      tcgetattr( PortDsc, &oldtio ); // save current port settings
      // set new port settings for non-canonical input processing
      newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
      newtio.c_iflag = IGNPAR; //  | IGNCR
      newtio.c_oflag = 0;
      newtio.c_lflag = 0;
      newtio.c_cc[VMIN]=0;
      newtio.c_cc[VTIME]=0;
      tcflush( PortDsc, TCIFLUSH );
      tcsetattr( PortDsc, TCSANOW, &newtio );

      fOpened  = true;
      nRes     = 0;
      LOG( LOG_INFO, "Comm port %s is opened", pszPortName );
    }
    else
    {
      LOG( LOG_ERROR, "Failed to set signal handler for Comm port");
    }
  }
  else
  {
    LOG( LOG_ERROR, "Failed to open Comm port %s", pszPortName );
  }

  if( nPortNum < 9 )
  {
    nPortNum++;
  }
  else
  {
    nPortNum = 0;
    // nRes = -2;
  }

  return nRes;
}


void TWDConnect::ProcessData( TWD_CommData * CommData )
{
  LOG( LOG_DEBUG, "Processing Data from WDG" );

  if( CommData->bSyncByte == UART_SYNC_BYTE )
  {
    switch( CommData->bCmd )
    {
      case WD_CMD_INFO:
        LOG( LOG_DEBUG, "Watchdog Info received" );
        if( GlobalData != NULL )
        {
          GlobalData->RunTime( CommData->Data.Info.dwRunTime );
          GlobalData->WDGSWVersion( CommData->Data.Info.SWVersion.value );
          GlobalData->LastTriggerTime( CommData->Data.Info.dwLastTriggerTime );
          GlobalData->ResetTime( CommData->Data.Info.dwResetTime );
          GlobalData->ResetsCount( CommData->Data.Info.dwResetsCount );
          GlobalData->TriggerCount( CommData->Data.Info.dwTriggerCount );
          GlobalData->Beep( CommData->Data.Info.bBeeping );
          GlobalData->StopReset( CommData->Data.Info.bStopReset );
          GlobalData->ResetLogEntryCount( CommData->Data.Info.dwLogEntryCount );

          AddTask( TSK_ANSWER_TASK, AT_INFO );
        }
      break;
      case WD_CMD_TRIGGER:
        LOG( LOG_DEBUG, "Watchdog Triggered" );

      break;
      case WD_CMD_SETINGS:
        AddTask( TSK_ANSWER_TASK, AT_SETTINGS );
      break;
      case WD_CMD_LOG:
        LOG( LOG_DEBUG, "Log entry %lu from WDG", CommData->Data.LogEntry.dwIndex );
        if( Watchdog != NULL )
        {
          Watchdog->SetWDGLogEntry( &CommData->Data.LogEntry );
          AddTask( TSK_ANSWER_TASK, AT_LOG_ENTRY );
        }

      break;
      case WD_CMD_CLEAR_LOG:
        LOG( LOG_DEBUG, "LOG CLEARED" );
        AddTask( TSK_ANSWER_TASK, AT_LOG_CLEAR );
      break;

      default:
        LOG( LOG_DEBUG, "Unknown CMD from WDG" );
      break;
    }
  }
}

bool TWDConnect::SendRequest( TCommCmds Cmd, void * vData )
{
  bool fRes = true;

  bool bSendIt = false;
  u08 bBuff[sizeof(TWD_CommData) + 3];
  TWD_CommData * pCommData = (TWD_CommData *)bBuff;
  u16convert wCRC;
  TWDGSettings * pWDGSettingsData;

  time_t rawtime;
  struct tm * timeinfo;


  if( fOpened )
  {
    pCommData->bSyncByte = UART_SYNC_BYTE;
    pCommData->bCmd = Cmd;

    time ( &rawtime );
    timeinfo = localtime ( &rawtime );

    pCommData->WDGDateTime.bSec   = timeinfo->tm_sec;
    pCommData->WDGDateTime.bMin   = timeinfo->tm_min;
    pCommData->WDGDateTime.bHour  = timeinfo->tm_hour;
    pCommData->WDGDateTime.bDay   = timeinfo->tm_mday - 1;
    pCommData->WDGDateTime.bMonth = timeinfo->tm_mon;
    pCommData->WDGDateTime.bYear  = (timeinfo->tm_year + 1900) - 2000;

    switch( Cmd )
    {
      case WD_CMD_TRIGGER:
      case WD_CMD_INFO:
        bSendIt = true;
      break;
      case WD_CMD_SETINGS:
        bSendIt = true;
        pWDGSettingsData = (TWDGSettings *)vData;

        LOG( LOG_DEBUG, "Sending Settings Cmd to WDG" );

        if( pWDGSettingsData->bSettingsCmd == WDGS_SET_SETTINGS )
        {
          pCommData->Data.Settings.bSettingsCmd = WD_CMD_SET_SETTINGS;

          pCommData->Data.Settings.dwSettings = 0;
          if( pWDGSettingsData->dwSettings & WDG_SET_MASK_RESET_TIME )
          {
            pCommData->Data.Settings.dwSettings |= WDG_SET_MASK_RESET_TIME;
            pCommData->Data.Settings.dwResetTime = pWDGSettingsData->dwResetTime;
          }

          if( pWDGSettingsData->dwSettings & WDG_SET_MASK_BEEP )
          {
            pCommData->Data.Settings.dwSettings |= WDG_SET_MASK_BEEP;
            pCommData->Data.Settings.bBeep = pWDGSettingsData->bBeep;
          }

          if( pWDGSettingsData->dwSettings & WDG_SET_MASK_STOP_RESET )
          {
            pCommData->Data.Settings.dwSettings |= WDG_SET_MASK_STOP_RESET;
            pCommData->Data.Settings.bStopReset = pWDGSettingsData->bStopReset;
          }

          if( pWDGSettingsData->dwSettings & WDG_SET_MASK_RESET_IT )
          {
            pCommData->Data.Settings.dwSettings |= WDG_SET_MASK_RESET_IT;
            pCommData->Data.Settings.bResetIt = pWDGSettingsData->bResetIt;
          }

          if( pWDGSettingsData->dwSettings & WDG_SET_MASK_CLEAR_CNT )
          {
            pCommData->Data.Settings.dwSettings |= WDG_SET_MASK_CLEAR_CNT;
            pCommData->Data.Settings.bClearCounters = pWDGSettingsData->bClearCounters;
          }

        }

      break;
      case WD_CMD_LOG:
        LOG( LOG_DEBUG, "Sending Get Log Item Cmd to WDG" );
        {
          bSendIt = true;
          TWD_LogEntryClient * LE = (TWD_LogEntryClient *)vData;
          pCommData->Data.LogEntry.dwIndex = LE->dwIndex;
        }
      break;

      case WD_CMD_CLEAR_LOG:
        LOG( LOG_DEBUG, "Sending CLEAR Log Cmd to WDG" );
        bSendIt = true;
      break;

      default:
        bSendIt = false;
      break;
    }

    if( bSendIt )
    {
      wCRC.value = CalculateCRC( bBuff, sizeof( TWD_CommData ) );
      bBuff[ sizeof( TWD_CommData )    ] = wCRC.bytes.low;
      bBuff[ sizeof( TWD_CommData ) + 1] = wCRC.bytes.high;

      int BytesToWrite = sizeof( TWD_CommData ) + 2;

      while( BytesToWrite > 0 )
      {
        int res = write( gsPortDsc, &bBuff, BytesToWrite );
        if( res >= 0 )
        {
          BytesToWrite -= res;
          if( Timer != NULL )
          {
            Timer->Load( SWT_ANSWER_DELAY_TIMER, pGlobalConfig->GetAnswerDelay() );
          }
        }
        else
        {
          LOG( LOG_ERROR, "Error on writing comm Port" );
          BytesToWrite = 0;
          fRes = false;
        }
      }

    }
  }
  return fRes;
}


#endif /* */
/*********************************** EOF *********************************/
